"""Insertion Devices Feedforward Devices."""

from ..namesys import SiriusPVName as _SiriusPVName
from ..search import IDSearch as _IDSearch
from ..idff.config import IDFFConfig as _IDFFConfig

from .device import Device as _Device, Devices as _Devices
from .pwrsupply import PowerSupplyFBP as _PowerSupplyFBP
from .ids import WIG as _WIG, APU as _APU, PAPU as _PAPU, EPU as _EPU


class IDFF(_Devices):
    """Insertion Device Feedforward Device."""

    class DEVICES(_WIG.DEVICES, _PAPU.DEVICES, _EPU.DEVICES):
        """."""
        ALL = _WIG.DEVICES.ALL + _PAPU.DEVICES.ALL + _EPU.DEVICES.ALL

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in IDFF.DEVICES.ALL:
            raise NotImplementedError(devname)

        self._devname = devname  # needed for _create_devices
        self._idffconfig = _IDFFConfig()

        self._pparametername = \
            _IDSearch.conv_idname_2_pparameter_propty(devname)

        self._kparametername = \
            _IDSearch.conv_idname_2_kparameter_propty(devname)

        self._devpp, self._devkp, self._devsch, self._devscv, self._devsqs = \
            self._create_devices(devname)

        # call base class constructor
        devices = [self._devkp, ]
        devices += self._devsch
        devices += self._devscv
        devices += self._devsqs
        super().__init__(devname=devname, devices=devices)

    @property
    def chnames(self):
        """Return CH corrector power supply names."""
        return _IDSearch.conv_idname_2_idff_chnames(self.devname)

    @property
    def cvnames(self):
        """Return CV corrector power supply names."""
        return _IDSearch.conv_idname_2_idff_cvnames(self.devname)

    @property
    def qsnames(self):
        """Return QS corrector power supply names."""
        return _IDSearch.conv_idname_2_idff_qsnames(self.devname)

    @property
    def chdevs(self):
        """Return CH corrector power supply devices."""
        return self._devsch

    @property
    def cvdevs(self):
        """Return CV corrector power supply devices."""
        return self._devscv

    @property
    def qsdevs(self):
        """Return QS corrector power supply names."""
        return self._devsqs

    @property
    def pparametername(self):
        """Return corresponding to ID pparameter."""
        return self._pparametername

    @property
    def kparametername(self):
        """Return corresponding to ID kparameter."""
        return self._kparametername

    @property
    def polarizations(self):
        """Return list of possible light polarizations for the ID."""
        return _IDSearch.conv_idname_2_polarizations(self.devname)

    @property
    def pparameter_mon(self):
        """Return pparameter value."""
        return self._devpp[self._pparametername]

    @property
    def kparameter_mon(self):
        """Return kparameter value."""
        return self._devkp[self._kparametername]

    @property
    def idffconfig(self):
        """."""
        return self._idffconfig

    def find_configs(self):
        """Find si_idff configurations in configdb."""
        return self._idffconfig.configdbclient.find_configs()

    def load_config(self, name):
        """Load IDFF configuration."""
        self._idffconfig.name = name
        if self._idffconfig.name == name:
            self._idffconfig.load()
        else:
            raise ValueError('Could not load configuration.')

    def calculate_setpoints(
            self, pparameter_value=None, kparameter_value=None):
        """Return correctors setpoints for a particular ID config.

        polarization - a string defining the required polarization for
        setpoint calculation.
        """
        polarization, pparameter_value, kparameter_value = \
            self.get_polarization_state(
                pparameter_value=pparameter_value,
                kparameter_value=kparameter_value)

        if not self._idffconfig:
            ValueError('IDFFConfig is not loaded!')

        if polarization not in self.idffconfig.polarizations:
            raise ValueError('Polarization is not compatible with ID.')
        if kparameter_value is None:
            kparameter_value = self.kparameter_mon
        setpoints = self.idffconfig.calculate_setpoints(
            polarization, kparameter_value)
        return setpoints, polarization, pparameter_value, kparameter_value

    def implement_setpoints(
            self, setpoints=None, corrdevs=None):
        """Implement setpoints in correctors."""
        if setpoints is None:
            setpoints, polarization, pparameter_value, kparameter_value = \
                self.calculate_setpoints(
                    pparameter_value=None,
                    kparameter_value=None)
        if corrdevs is None:
            corrdevs = self._devsch + self._devscv + self._devsqs
        for pvname, value in setpoints.items():
            # find corrdev corresponding to pvname
            for dev in corrdevs:
                if dev.devname in pvname:
                    pvname = _SiriusPVName(pvname)
                    dev[pvname.propty] = value
                    break
        return polarization, pparameter_value, kparameter_value

    def check_valid_value(self, value):
        """Check consistency of SI_IDFF configuration."""
        if not super().check_valid_value(value):
            raise ValueError('Value incompatible with config template')

        configs = value['polarizations']
        pvnames = {key: value for key, value in value['pvnames'] \
            if key not in ('pparameters', 'kparameters')}
        corrlabels = set(pvnames.keys())

        # check pvnames in configs
        pvsconfig = set(pvnames.values())
        pvsidsearch = set(self.chnames + self.cvnames + self.qsnames)
        symm_diff = pvsconfig ^ pvsidsearch
        if symm_diff:
            raise ValueError('List of pvnames in config is not consistent')

        # check polarizations in configs
        pconfig = set(configs.keys())
        pidsearch = set(_IDSearch.conv_idname_2_polarizations(self.devname))
        symm_diff = pconfig ^ pidsearch
        if symm_diff:
            raise ValueError(
                'List of polarizations in config is not consistent')

        # check polarization tables consistency
        for polarization, table in configs.items():
            corrtable = {key: value for key, value in table \
                if key not in ('pparameters', 'kparameters')}

            # check 'pparameter'
            if 'pparameter' not in table:
                raise ValueError(
                    'Missing pparameter in polarization configuration.')

            # check 'kparameter'
            if 'kparameter' not in table:
                raise ValueError(
                    'Missing kparameter in polarization configuration.')

            # check corr label list
            corrlabels_config = set(corrtable.keys())
            symm_diff = corrlabels ^ corrlabels_config
            if symm_diff:
                raise ValueError(
                    'List of corrlabels in config is not consistent')

            # check nrpts in tables
            param = 'pparameter' if polarization == 'none' else 'kparameter'
            nrpts_corrtables = set([len(table) for table in corrtable.values()])
            nrpts_kparameter = set([len(table[param]), ])
            symm_diff = nrpts_corrtables ^ nrpts_kparameter
            if symm_diff:
                msg = (
                    'Corrector tables and kparameter list in config'
                    ' are not consistent')
                raise ValueError(msg)

        return True

    def get_polarization_state(
            self, pparameter_value=None, kparameter_value=None):
        """."""
        if self.pparametername is None:
            if kparameter_value is None:
                kparameter_value = self.kparameter_mon
            return self.polarizations[0], pparameter_value, kparameter_value
        if pparameter_value:
            pparameter_value = self.pparameter_mon
        if kparameter_value is None:
            kparameter_value = self.kparameter_mon
        if None in (pparameter_value, kparameter_value):
            return None, pparameter_value, kparameter_value
        polarization = self.idffconfig.get_polarization_state(
            pparameter=pparameter_value, kparameter=kparameter_value)
        return polarization, pparameter_value, kparameter_value

    def _create_devices(self, devname):
        param_auto_mon = False
        devpp = _Device(
            devname=devname,
            properties=(self._pparametername, ),
            auto_monitor_mon=param_auto_mon)
        devkp = _Device(
            devname=devname,
            properties=(self._kparametername, ),
            auto_monitor_mon=param_auto_mon)
        devsch = [_PowerSupplyFBP(devname=dev) for dev in self.chnames]
        devscv = [_PowerSupplyFBP(devname=dev) for dev in self.cvnames]
        devsqs = [_PowerSupplyFBP(devname=dev) for dev in self.qsnames]

        return devpp, devkp, devsch, devscv, devsqs


class WIGIDFF(IDFF):
    """Wiggler Feedforward."""

    class DEVICES(_WIG.DEVICES):
        """."""

    @property
    def gap_mon(self):
        """."""
        return self.kparameter_mon


class PAPUIDFF(IDFF):
    """PAPU Feedforward."""

    class DEVICES(_PAPU.DEVICES):
        """."""

    @property
    def phase_mon(self):
        """."""
        return self.pparameter_mon


class EPUIDFF(IDFF):
    """EPU Feedforward."""

    class DEVICES(_EPU.DEVICES):
        """."""

    @property
    def phase_mon(self):
        """."""
        return self.pparameter_mon

    @property
    def gap_mon(self):
        """."""
        return self.kparameter_mon


class APUIDFF(_Devices):
    """APU Feedforward."""

    class DEVICES(_APU.DEVICES):
        """."""

    @property
    def phase_mon(self):
        """."""
        return self.kparameter_mon



#!/usr/bin/env python-sirius

"""Module to test AS-AP-ChromCorr Soft IOC main module."""

import unittest
from unittest import mock
import siriuspy.util as util
from siriuspy.opticscorr.csdev import Const
from siriuspy.opticscorr.chrom import ChromCorrApp


PUB_INTERFACE = (
    'update_corrparams_pvs',
    'process',
    'set_chrom_x',
    'set_chrom_y',
    'set_deltachrom_x',
    'set_deltachrom_y',
    'set_meas_chrom_dfreq_rf',
    'set_meas_chrom_wait_tune',
    'set_meas_chrom_nrsteps',
    'cmd_meas_chrom',
    'set_meas_config_dsl_sf',
    'set_meas_config_dsl_sd',
)


class TestASAPChromCorrMain(unittest.TestCase):
    """Test AS-AP-ChromCorr Soft IOC."""

    def setUp(self):
        """Initialize Soft IOC."""
        self.q_ok = {
            'matrix': [
                [1.2105, 0.5759, 0.3767, 0.6833, 0.2642,
                 2.4732, 1.154, 0.8185, 1.3816, 0.4859,
                 1.2563, 0.5672, 0.4101, 0.7106, 0.2702],
                [-1.0285, -0.3025, -1.1841, -1.2897, -0.6505,
                 -2.0764, -0.6426, -2.372, -2.6034, -1.2536,
                 -1.0478, -0.3373, -1.1728, -1.2834, -0.6496]],
            'nominal SLs': [
                28.7743, 22.6154, -24.448, -13.3281, -20.9911,
                34.1822, 29.6731, -21.2454, -18.3342, -26.0719,
                34.3874, 29.7755, -21.3459, -18.3422, -26.1236],
            'nominal chrom': [2.5756, 2.5033]}
        self.sfams = ['SFA1', 'SFA2', 'SDA1', 'SDA2', 'SDA3',
                      'SFB1', 'SFB2', 'SDB1', 'SDB2', 'SDB3',
                      'SFP1', 'SFP2', 'SDP1', 'SDP2', 'SDP3']
        cs_patcher = mock.patch(
            "siriuspy.opticscorr.base._ConfigDBClient", autospec=True)
        self.addCleanup(cs_patcher.stop)
        self.mock_cs = cs_patcher.start()
        self.mock_cs.return_value.get_config_value.return_value = self.q_ok
        pv_patcher = mock.patch(
            "siriuspy.opticscorr.base._PV", autospec=True)
        self.addCleanup(pv_patcher.stop)
        self.mock_pv = pv_patcher.start()
        cnh_patcher = mock.patch(
            "siriuspy.opticscorr.base._HandleConfigNameFile",
            autospec=True)
        self.addCleanup(cnh_patcher.stop)
        self.mock_cnh = cnh_patcher.start()
        self.mock_cnh.return_value.get_config_name.return_value = \
            'SI.V24.04_S05.01'
        self.app = ChromCorrApp('SI')

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            ChromCorrApp, PUB_INTERFACE, print_flag=True)
        self.assertTrue(valid)

    def test_write_ok_syncoff_Apply(self):
        """Test write on ApplyDelta-Cmd in normal operation, sync mode off."""
        self.app._sync_corr = 0

        self.app._status = 0
        self.assertFalse(self.app.write('ApplyDelta-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, len(self.sfams))

        self.app._status = 0b10000
        self.assertFalse(self.app.write('ApplyDelta-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, 2*len(self.sfams))

    def test_write_ok_syncon_Apply(self):
        """Test write on ApplyDelta-Cmd in normal operation, sync mode on."""
        self.app._sync_corr = 1
        self.app._status = 0
        self.assertFalse(self.app.write('ApplyDelta-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, 1+len(self.sfams))

    def test_write_err_syncon_Apply(self):
        """Test write on ApplyDelta-Cmd on status error."""
        self.app._sync_corr = Const.SyncCorr.On
        self.app._status = 0b10000
        self.assertFalse(self.app.write('ApplyDelta-Cmd', 0))
        self.mock_pv.return_value.put.assert_not_called()

    def test_write_CorrMeth(self):
        """Test write on CorrMeth-Sel."""
        self.assertTrue(self.app.write('CorrMeth-Sel', 1))
        self.assertEqual(self.app._corr_method, 1)

    def test_write_SyncCorr(self):
        """Test write on SyncCorr-Sel."""
        self.mock_pv.return_value.connected = True
        self.assertTrue(self.app.write('SyncCorr-Sel', 1))
        self.assertEqual(self.app._sync_corr, 1)

    def test_write_ok_ConfigPS(self):
        """Test write on ConfigPS-Cmd in normal operation."""
        self.mock_pv.return_value.connected = True
        self.assertFalse(self.app.write('ConfigPS-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, 2*len(self.sfams))

    def test_write_connerror_ConfigPS(self):
        """Test write on ConfigPS-Cmd on connection error."""
        self.mock_pv.return_value.connected = False
        self.assertFalse(self.app.write('ConfigPS-Cmd', 0))
        self.mock_pv.return_value.put.assert_not_called()

    def test_write_ok_ConfigTI(self):
        """Test write on ConfigTiming-Cmd in normal operation."""
        self.mock_pv.return_value.connected = True
        self.assertFalse(self.app.write('ConfigTiming-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, 9)

    def test_write_connerror_ConfigTI(self):
        """Test write on ConfigTiming-Cmd in connection errorn."""
        self.mock_pv.return_value.connected = False
        self.assertFalse(self.app.write('ConfigTiming-Cmd', 0))
        self.mock_pv.return_value.put.assert_not_called()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python-sirius

"""Module to test AS-AP-CurrInfo Current Soft IOC main module."""

import unittest
from unittest import mock
import siriuspy.util as util
from siriuspy.currinfo.csdev import Const
from siriuspy.currinfo import SICurrInfoApp
from siriuspy.currinfo.main import _CurrInfoApp


PUB_INTERFACE_BASE = (
    'pvs_database',
    'init_database',
    'process',
    'read',
    'write',
    'close',
)

PUB_INTERFACE_SI = (
    'HARMNUM',
    'HARMNUM_RATIO',
    'CURR_THRESHOLD',
    'MAX_CURRENT',
    'init_database',
    'read',
    'write',
)


class TestASAPCurrInfoCurrentMain(unittest.TestCase):
    """Test AS-AP-CurrInfo Soft IOC."""

    def setUp(self):
        """Set Up tests."""
        ca_patcher = mock.patch(
            "siriuspy.currinfo.main._ClientArch", autospec=True)
        self.addCleanup(ca_patcher.stop)
        self.mock_ca = ca_patcher.start()
        self.mock_ca.return_value.getData.return_value = None
        pv_patcher = mock.patch(
            "siriuspy.currinfo.main._PV", autospec=True)
        self.addCleanup(pv_patcher.stop)
        self.mock_pv = pv_patcher.start()
        self.mock_pv.return_value.connected = False
        self.app = SICurrInfoApp()

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            _CurrInfoApp, PUB_INTERFACE_BASE, print_flag=True)
        self.assertTrue(valid)
        valid = util.check_public_interface_namespace(
            SICurrInfoApp, PUB_INTERFACE_SI, print_flag=True)
        self.assertTrue(valid)

    def test_write_DCCTFltCheck(self):
        """Test write DCCTFltCheck-Sel."""
        self.app.write(
            'SI-Glob:AP-CurrInfo:DCCTFltCheck-Sel', Const.DCCTFltCheck.On)
        self.assertEqual(self.app._dcctfltcheck_mode, Const.DCCTFltCheck.On)

    def test_write_DCCT_FltCheckOn(self):
        """Test write DCCT-Sel."""
        self.app._dcctfltcheck_mode = Const.DCCTFltCheck.On
        init_status = self.app._dcct_mode
        self.app.write('SI-Glob:AP-CurrInfo:DCCT-Sel', Const.DCCT.DCCT13C4)
        self.app.write('SI-Glob:AP-CurrInfo:DCCT-Sel', Const.DCCT.DCCT14C4)
        end_status = self.app._dcct_mode
        self.assertEqual(init_status, end_status)

    def test_write_DCCT_FltCheckOff(self):
        """Test write DCCT-Sel."""
        self.app.write(
            'SI-Glob:AP-CurrInfo:DCCTFltCheck-Sel', Const.DCCTFltCheck.Off)
        self.app.write('SI-Glob:AP-CurrInfo:DCCT-Sel', Const.DCCT.DCCT13C4)
        self.assertEqual(self.app._dcct_mode, Const.DCCT.DCCT13C4)


if __name__ == "__main__":
    unittest.main()



"""Linac Diag App."""

from ... import csdev as _csdev


class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    DIAG_STATUS_LABELS_RF = (
        'Disconnected',
        'State Off',
        'Trigger Off',
        'Integral Off',
        'Feedback Off',
        'Amp-(SP|RB) are different',
        'Phase-(SP|RB) are different',
        'IxQ (SP|Mon) are different',
    )

    DIAG_STATUS_LABELS_PU = (
        'Disconnected',
        'Run/Stop not ok',
        'PreHeat not ok',
        'Charge_Allowed not ok',
        'TrigOut_Allowed not ok',
        'Emer_Stop not ok',
        'CPS_ALL not ok',
        'Thy_Heat not ok',
        'Kly_Heat not ok',
        'LV_Rdy_OK not ok',
        'HV_Rdy_OK not ok',
        'TRIG_Rdy_OK not ok',
        'MOD_Self_Fault not ok',
        'MOD_Sys_Ready not ok',
        'TRIG_Norm not ok',
        'Pulse_Current not ok',
        'Voltage-(SP|RB) are different',
        'Current-(SP|RB) are different',
        'CHARGE is off',
        'TRIGOUT is off',
    )

    DIAG_STATUS_LABELS_EG_HVPS = (
        'Disconnected',
        'Swicth Status Off',
        'Enable Status Off',
        'Voltage (SP|Mon) are different',
    )

    DIAG_STATUS_LABELS_EG_FILA = (
        'Disconnected',
        'Swicth Status Off',
        'Current-(SP|Mon) are different',
    )


_et = ETypes  # syntactic sugar


class Const(_csdev.Const):
    """Local constant types."""

    SHB = 'LI-01:RF-SHB'
    KLY1 = 'LI-01:RF-Kly-1'
    KLY2 = 'LI-01:RF-Kly-2'
    RF_DEVICES = [SHB, KLY1, KLY2]
    MOD1 = 'LI-01:PU-Modltr-1'
    MOD2 = 'LI-01:PU-Modltr-2'
    PU_DEVICES = [MOD1, MOD2]
    HVPS = 'LI-01:EG-HVPS'
    FILA = 'LI-01:EG-FilaPS'
    DEV_2_LINAME = {
        HVPS: 'LI-01:EG-HVPS',
        FILA: 'LI-01:EG-FilaPS',
        SHB: 'LA-RF:LLRF:BUN1',
        KLY1: 'LA-RF:LLRF:KLY1',
        KLY2: 'LA-RF:LLRF:KLY2',
        MOD1: 'LI-01:PU-Modltr-1',
        MOD2: 'LI-01:PU-Modltr-2'}
    ALL_DEVICES = DEV_2_LINAME.keys()

    LI_RF_AMP_TOL = 1e-2
    LI_RF_PHS_TOL = 1e-2
    LI_RF_IxQ_TOL = 1e-2
    LI_PU_VOLT_TOL = 1e-1
    LI_PU_CURR_TOL = 1e-1
    LI_HVPS_TOL = 1.5
    LI_FILAPS_TOL = 1e-1


_c = Const  # syntactic sugar


def conv_dev_2_liname(device):
    """Return Linac pvname."""
    return _c.DEV_2_LINAME[device]


def get_li_diag_status_labels(device):
    """Return Diag Status Labels enum."""
    if 'RF' in device:
        return _et.DIAG_STATUS_LABELS_RF
    if 'PU' in device:
        return _et.DIAG_STATUS_LABELS_PU
    if 'HVPS' in device:
        return _et.DIAG_STATUS_LABELS_EG_HVPS
    if 'FilaPS' in device:
        return _et.DIAG_STATUS_LABELS_EG_FILA
    raise ValueError('Labels not defined to '+device+'.')


def get_li_diag_propty_database(device):
    """Return property database of diagnostics for linac devices."""
    enums = get_li_diag_status_labels(device)
    dbase = {
        'DiagVersion-Cte': {'type': 'str', 'value': 'UNDEF'},
        'DiagStatus-Mon': {'type': 'int', 'value': 0,
                           'hilim': 1, 'hihi': 1, 'high': 1,
                           'low': -1, 'lolo': -1, 'lolim': -1
                           },
        'DiagStatusLabels-Cte': {'type': 'string', 'count': len(enums),
                                 'value': enums}
    }
    if 'RF' in device:
        atol = _c.LI_RF_AMP_TOL
        ptol = _c.LI_RF_PHS_TOL
        vtol = _c.LI_RF_IxQ_TOL
        dbase.update({
            'DiagAmpDiff-Mon': {
                'type': 'float', 'value': 0.0,
                'hilim': atol, 'hihi': atol, 'high': atol,
                'low': -atol, 'lolo': -atol, 'lolim': -atol},
            'DiagPhaseDiff-Mon': {
                'type': 'float', 'value': 0.0,
                'hilim': ptol, 'hihi': ptol, 'high': ptol,
                'low': -ptol, 'lolo': -ptol, 'lolim': -ptol},
            'DiagIxQDiff-Mon': {
                'type': 'float', 'value': 0.0,
                'hilim': vtol, 'hihi': vtol, 'high': vtol,
                'low': -vtol, 'lolo': -vtol, 'lolim': -vtol},
        })
    elif 'PU' in device:
        vtol = _c.LI_PU_VOLT_TOL
        ctol = _c.LI_PU_CURR_TOL
        dbase.update({
            'DiagVoltageDiff-Mon': {
                'type': 'float', 'value': 0.0,
                'hilim': vtol, 'hihi': vtol, 'high': vtol,
                'low': -vtol, 'lolo': -vtol, 'lolim': -vtol},
            'DiagCurrentDiff-Mon': {
                'type': 'float', 'value': 0.0,
                'hilim': ctol, 'hihi': ctol, 'high': ctol,
                'low': -ctol, 'lolo': -ctol, 'lolim': -ctol},
        })
    elif 'HVPS' in device:
        dtol = _c.LI_HVPS_TOL
        dbase.update({
            'DiagVoltDiff-Mon': {
                'type': 'float', 'value': 0.0,
                'hilim': dtol, 'hihi': dtol, 'high': dtol,
                'low': -dtol, 'lolo': -dtol, 'lolim': -dtol}})
    elif 'FilaPS' in device:
        dtol = _c.LI_FILAPS_TOL
        dbase.update({
            'DiagCurrentDiff-Mon': {
                'type': 'float', 'value': 0.0,
                'hilim': dtol, 'hihi': dtol, 'high': dtol,
                'low': -dtol, 'lolo': -dtol, 'lolim': -dtol}})
    dbase = _csdev.add_pvslist_cte(dbase, 'Diag')
    return dbase



"""TB posang correction configurations.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return a dict with ramp settings."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }

    return _dict


# Position-Angle Correction Response Matrices for the Booster Injection
# (TB correctors)
#
#
# Horizontal Matrix (ch-sept):
#
#  | DeltaPosX @ TB-04:PM-InjSept |    | h11  h12 |   |  Kick TB-04:PS-CH-1   |
#  |                              | =  |          | * |                       |
#  | DeltaAngX @ TB-04:PM-InjSept |    | h21  h22 |   | Kick TB-04:PM-InjSept |
#
# Horizontal Matrix (ch-ch):
#
#  | DeltaPosX @ TB-04:PM-InjSept |    | h11  h12 |   |  Kick TB-04:PS-CH-1  |
#  |                              | =  |          | * |                      |
#  | DeltaAngX @ TB-04:PM-InjSept |    | h21  h22 |   |  Kick TB-04:PS-CH-2  |
#
# Data structure:
#         h11   h12
#         h21   h22
#
#
# Vertical Matrix:
#
#  | DeltaPosY @ TB-04:PM-InjSept |    | v11  v12 |   | Kick TB-04:PS-CV-1 |
#  |                              | =  |          | * |                    |
#  | DeltaAngY @ TB-04:PM-InjSept |    | v21  v22 |   | Kick TB-04:PS-CV-2 |
#
# Data structure:
#         v11   v12
#         v21   v22

_template_dict = {
    'respm-x': [[0.0, 0.0],
                [0.0, 0.0]],
    'respm-y': [[0.0, 0.0],
                [0.0, 0.0]],
}



"""Machine Shift Control System App."""

from .. import csdev as _csdev


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    MACHSHIFT = (
        'Users', 'Commissioning', 'Conditioning',
        'Injection', 'MachineStudy', 'Maintenance',
        'Standby', 'Shutdown', 'MachineStartup')
    PROGMD_USERS_SHIFT = ('No', 'Yes')


_et = ETypes


# --- Const class ---

class Const(_csdev.Const):
    """Const class."""

    MachShift = _csdev.Const.register('MachShift', _et.MACHSHIFT)
    ProgmdUsersShift = _csdev.Const.register(
        'ProgmdUsersShift', _et.PROGMD_USERS_SHIFT)


_c = Const


# --- Databases ---


def get_machshift_propty_database():
    """Return property database of machine shift IOC."""
    dbase = {
        'Version-Cte': {'type': 'str', 'value': 'UNDEF'},
        'Mode-Sel': {
            'type': 'enum', 'value': _c.MachShift.Commissioning,
            'enums': _et.MACHSHIFT},
        'Mode-Sts': {
            'type': 'enum', 'value': _c.MachShift.Commissioning,
            'enums': _et.MACHSHIFT},
        'ProgmdUsersShift-Mon': {
            'type': 'enum', 'value': _c.ProgmdUsersShift.No,
            'enums': _et.PROGMD_USERS_SHIFT, 'scan': 60}
    }
    dbase = _csdev.add_pvslist_cte(dbase)
    return dbase



"""Time conversion module."""

from . import exceptions as _exceptions
from datetime import datetime as _datetime, timedelta as _timedelta


class Time(_datetime):
    """Time conversion class.

    Usage options:
        Time(timestamp)
            timestamp is a float/int keyword/positional argument.

        Time(timestamp_string)
        Time(timestamp_string, timestamp_format='%Y-%m-%d %H:%M:%S.%f')
            timestamp_string is a str keyword/positional argument.
            timestamp_format is an optional keyword argument for string
                formating. Defaults to '%Y-%m-%d %H:%M:%S.%f'.

        Time(year, month, day)
        Time(year, month, day, hour)
        Time(year, month, day, hour, minute)
        Time(year, month, day, hour, minute, second)
        Time(year, month, day, hour, minute, second, microsecond)
        Time(year, month, day, hour, minute, second, microsecond, tzinfo)
            year, month, day, hour, minute, second, microsecond
                are integer keyword/positional arguments.
            tzinfo must be None or of a tzinfo subclass keyword/positional
                argument.
    """

    _DEFAULT_TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
    _DATETIME_ARGS = {'year', 'month', 'day', 'hour', 'minute',
                      'second', 'microsecond', 'tzinfo'}

    def __new__(cls, *args, **kwargs):
        """New object."""
        if not args and not kwargs:
            raise _exceptions.TypeError(
                'no arguments found to build Time object')
        if len(args) == 1:
            if isinstance(args[0], (float, int)):
                return Time.fromtimestamp(args[0])
            if isinstance(args[0], str):
                timestamp_format = \
                    kwargs['timestamp_format'] if 'timestamp_format'\
                    in kwargs else Time._DEFAULT_TIMESTAMP_FORMAT
                return Time.strptime(args[0], timestamp_format)
            raise _exceptions.TypeError(
                f'argument of unexpected type {type(args[0])}')
        if len(kwargs) == 1:
            if 'timestamp' in kwargs:
                return Time.fromtimestamp(kwargs['timestamp'])
            if 'timestamp_string' in kwargs:
                return Time.strptime(
                    kwargs['timestamp_string'], Time._DEFAULT_TIMESTAMP_FORMAT)
            if set(kwargs.keys()) & Time._DATETIME_ARGS:
                raise _exceptions.TypeError(
                    'missing input arguments, verify usage options.')
            raise _exceptions.TypeError(
                f'unexpected key argument {kwargs}')
        if len(kwargs) == 2:
            if set(kwargs.keys()) == {'timestamp_string', 'timestamp_format'}:
                return Time.strptime(
                    kwargs['timestamp_string'], kwargs['timestamp_format'])
            if set(kwargs.keys()) & Time._DATETIME_ARGS:
                raise _exceptions.TypeError(
                    'missing input arguments, verify usage options.')
            raise _exceptions.TypeError(
                f'unexpected key arguments {list(kwargs.keys())}')
        return super().__new__(cls, *args, **kwargs)

    def get_iso8601(self):
        """Get iso8601 format."""
        return self.astimezone().isoformat()

    def __add__(self, other):
        """Addition."""
        if isinstance(other, _datetime):
            return super().__add__(other)
        if isinstance(other, (float, int)):
            add = super().__add__(_timedelta(seconds=other))
        else:
            add = super().__add__(other)
        return Time(timestamp=add.timestamp())

    def __sub__(self, other):
        """Subtraction."""
        if isinstance(other, _datetime):
            return super().__sub__(other)
        if isinstance(other, (float, int)):
            sub = super().__sub__(_timedelta(seconds=other))
        else:
            sub = super().__sub__(other)
        return Time(timestamp=sub.timestamp())


def get_time_intervals(
        time_start, time_stop, interval, return_isoformat=False):
    """Return intervals of 'interval' duration from time_start to time_stop."""
    if time_start + interval >= time_stop:
        timestamp_start = time_start.get_iso8601() \
            if return_isoformat else time_start
        timestamp_stop = time_stop.get_iso8601() \
            if return_isoformat else time_stop
    else:
        t_start = time_start
        t_stop = t_start + interval
        timestamp_start = [t_start, ]
        timestamp_stop = [t_stop, ]
        while t_stop < time_stop:
            t_start += interval
            t_stop = t_stop + interval
            if t_stop + interval > time_stop:
                t_stop = time_stop
            timestamp_start.append(t_start)
            timestamp_stop.append(t_stop)
        if return_isoformat:
            timestamp_start = [t.get_iso8601() for t in timestamp_start]
            timestamp_stop = [t.get_iso8601() for t in timestamp_stop]
    return timestamp_start, timestamp_stop



"""ID Search module."""

from ..namesys import Filter as _Filter
from ..namesys import SiriusPVName as _SiriusPVName


class IDSearch:
    """ID Search Class."""

    # NOTE: Maybe we should move these mappings to
    # static tables in our cs-constants server.

    # NOTE: see
    # https://wiki-sirius.lnls.br/mediawiki/index.php/Machine:Insertion_Devices

    _beamline2idname = {
        'CARNAUBA': 'SI-06SB:ID-APU22',  # titular: VPU19
        'CATERETE': 'SI-07SP:ID-APU22',  # titular: VPU19
        'EMA':      'SI-08SB:ID-APU22',  # titular: IVU18
        'MANACA':   'SI-09SA:ID-APU22',  # titular: 2 x APU22
        'SABIA':    'SI-10SB:ID-EPU50',  # titular: 2 x DELTA52
        'IPE':      'SI-11SP:ID-APU58',  # titular: 2 x APPLE-II
        'PAINEIRA': 'SI-14SB:ID-WIG180',  # titular: IVU18
        'SAPUCAIA': 'SI-17SA:ID-PAPU50',  # titular: 2 x APU22
    }

    _idname2beamline = {v: k for k, v in _beamline2idname.items()}

    _idname_2_idff = {
        'SI-06SB:ID-APU22': None,
        'SI-07SP:ID-APU22': None,
        'SI-08SB:ID-APU22': None,
        'SI-09SA:ID-APU22': None,
        'SI-10SB:ID-EPU50': {
            'polarizations': (
                'none', 'circularn', 'horizontal', 'circularp', 'vertical',
                ),
            'pparameter': 'SI-10SB:ID-EPU50:Phase-Mon',
            'kparameter': 'SI-10SB:ID-EPU50:Gap-Mon',
            'ch1': 'SI-10SB:PS-CH-1:Current-SP',  # upstream
            'ch2': 'SI-10SB:PS-CH-2:Current-SP',  # downstream
            'cv1': 'SI-10SB:PS-CV-1:Current-SP',
            'cv2': 'SI-10SB:PS-CV-2:Current-SP',
            'qs1': 'SI-10SB:PS-QS-1:Current-SP',
            'qs2': 'SI-10SB:PS-QS-2:Current-SP',
        },
        'SI-11SP:ID-APU58': None,
        'SI-14SB:ID-WIG180': {
            'polarizations': ('horizontal', ),
            'pparameter': None,
            'kparameter': 'SI-14SB:ID-WIG180:Gap-Mon',
            'ch1': 'SI-14SB:PS-CH-1:Current-SP',  # upstream
            'ch2': 'SI-14SB:PS-CH-2:Current-SP',  # downstream
        },
        'SI-17SA:ID-PAPU50': {
            'polarizations': ('horizontal', ),
            'pparameter': None,
            'kparameter': 'SI-17SA:ID-PAPU50:Phase-Mon',
            'ch1': 'SI-17SA:PS-CH-1:Current-SP',  # upstream
            'ch2': 'SI-17SA:PS-CH-2:Current-SP',  # downstream
            'cv1': 'SI-17SA:PS-CV-1:Current-SP',
            'cv2': 'SI-17SA:PS-CV-2:Current-SP',
        },
    }

    @staticmethod
    def get_idnames(filters=None):
        """Return a sorted and filtered list of all ID names."""
        idnames_list = list(IDSearch._idname2beamline.keys())
        idnames = _Filter.process_filters(idnames_list, filters=filters)
        return sorted(idnames)

    @staticmethod
    def get_beamlines(filters=None):
        """Return a sorted and filtered list of all ID beamlines."""
        beamlines_list = list(IDSearch._idname2beamline.values())
        beamlines = _Filter.process_filters(beamlines_list, filters=filters)
        return sorted(beamlines)

    @staticmethod
    def conv_idname_2_beamline(idname):
        """Return the beam line name corresponding to sector."""
        if idname in IDSearch._idname2beamline.keys():
            return IDSearch._idname2beamline[idname]
        return None

    @staticmethod
    def conv_beamline_2_idname(beamline):
        """Return the beam line name corresponding to idname."""
        if beamline in IDSearch._beamline2idname.keys():
            return IDSearch._beamline2idname[beamline]
        return None

    @staticmethod
    def get_idname_2_beamline_dict():
        """Return idname to beamline dictionary."""
        return IDSearch._idname2beamline.copy()

    @staticmethod
    def get_beamline_2_idname_dict():
        """Return beamline to idname dictionary."""
        return IDSearch._beamline2idname.copy()

    @staticmethod
    def conv_idname_2_idff(idname):
        """Return the IDFF dictionary for a given ID name."""
        return dict(IDSearch._idname_2_idff[idname])

    @staticmethod
    def conv_idname_2_pparameter_propty(idname):
        """."""
        idff = IDSearch.conv_idname_2_idff(idname)
        pparameter = idff['pparameter']
        if pparameter:
            pvname = _SiriusPVName(pparameter)
            return pvname.propty
        else:
            return None

    @staticmethod
    def conv_idname_2_kparameter_propty(idname):
        """."""
        idff = IDSearch.conv_idname_2_idff(idname)
        pvname = _SiriusPVName(idff['kparameter'])
        return pvname.propty

    @staticmethod
    def conv_idname_2_idff_chnames(idname):
        """."""
        return IDSearch._get_devname_from_idff(idname, ('ch1', 'ch2'))

    @staticmethod
    def conv_idname_2_idff_cvnames(idname):
        """."""
        return IDSearch._get_devname_from_idff(idname, ('cv1', 'cv2'))

    @staticmethod
    def conv_idname_2_idff_qsnames(idname):
        """."""
        return IDSearch._get_devname_from_idff(idname, ('qs1', 'qs2'))

    @staticmethod
    def conv_idname_2_polarizations(idname):
        """Return ID light polarizations."""
        polarizations = IDSearch._idname_2_idff[idname]['polarizations']
        return polarizations

    # --- private ----

    @staticmethod
    def _get_devname_from_idff(idname, correctors):
        idff = IDSearch.conv_idname_2_idff(idname)
        corrs = list()
        for corr in correctors:
            if corr in idff:
                pvname = _SiriusPVName(idff[corr])
                corrs.append(pvname.device_name)
        return corrs



"""Define Insertion Devices."""

import time as _time
import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from ..namesys import SiriusPVName as _SiriusPVName

from .device import Device as _Device


class APU(_Device):
    """APU Insertion Device."""

    class DEVICES:
        """."""

        APU22_06SB = 'SI-06SB:ID-APU22'
        APU22_07SP = 'SI-07SP:ID-APU22'
        APU22_08SB = 'SI-08SB:ID-APU22'
        APU22_09SA = 'SI-09SA:ID-APU22'
        APU58_11SP = 'SI-11SP:ID-APU58'
        ALL = (APU22_06SB, APU22_07SP, APU22_08SB, APU22_09SA,
               APU58_11SP, )

    _properties = (
        'DevCtrl-Cmd', 'Moving-Mon',
        'PhaseSpeed-SP', 'PhaseSpeed-Mon',
        'Phase-SP', 'Phase-Mon',
        'Kx-SP', 'Kx-Mon',
    )

    _DEF_TIMEOUT = 10  # [s]
    _CMD_MOVE = 3
    _MOVECHECK_SLEEP = 0.1  # [s]

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in APU.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, properties=APU._properties, auto_monitor_mon=True)

    @property
    def phase(self):
        """Return APU phase [mm]."""
        return self['Phase-Mon']

    @phase.setter
    def phase(self, value):
        """Set APU phase [mm]."""
        self['Phase-SP'] = value

    @property
    def phase_sp(self):
        """Return APU phase SP [mm]."""
        return self['Phase-SP']

    @property
    def phase_speed(self):
        """Return APU phase speed [mm/s]."""
        return self['PhaseSpeed-Mon']

    @phase_speed.setter
    def phase_speed(self, value):
        """Set APU phase_speed [mm/s]."""
        self['PhaseSpeed-SP'] = value

    @property
    def phase_speed_sp(self):
        """Return APU phase speed SP [mm/s]."""
        return self['PhaseSpeed-SP']

    @property
    def idkx(self):
        """Return APU Kx."""
        return self['Kx-SP']

    @idkx.setter
    def idkx(self, value):
        """Set APU Kx."""
        self['Kx-SP'] = value

    @property
    def is_moving(self):
        """Return True if phase is changing."""
        return round(self['Moving-Mon']) == 1

    def cmd_move(self, timeout=_DEF_TIMEOUT):
        """."""
        self['DevCtrl-Cmd'] = APU._CMD_MOVE
        return True

    def wait_move(self):
        """Wait for phase movement to complete."""
        _time.sleep(APU._MOVECHECK_SLEEP)
        while self.is_moving:
            _time.sleep(APU._MOVECHECK_SLEEP)


class PAPU(_Device):
    """PAPU Insertion Device."""

    class DEVICES:
        """."""

        PAPU50_17SA = 'SI-17SA:ID-PAPU50'
        ALL = (PAPU50_17SA, )

    TOLERANCE_PHASE = 0.01  # [mm]

    _SHORT_SHUT_EYE = 0.1  # [s]
    _default_timeout = 8  # [s]

    _properties_papu = (
        'Home-Cmd', 'EnblPwrPhase-Cmd', 'ClearErr-Cmd',
        'BeamLineCtrl-Mon',  # 'Home-Mon',
        )
    _properties = (
        'PeriodLength-Cte',
        'BeamLineCtrlEnbl-Sel', 'BeamLineCtrlEnbl-Sts',
        'Moving-Mon',
        'PwrPhase-Mon',
        'EnblAndReleasePhase-Sel', 'EnblAndReleasePhase-Sts',
        'AllowedToChangePhase-Mon',
        'ParkedPhase-Cte',
        'Phase-SP', 'Phase-RB', 'Phase-Mon',
        'PhaseSpeed-SP', 'PhaseSpeed-RB', 'PhaseSpeed-Mon',
        'MaxPhaseSpeed-SP', 'MaxPhaseSpeed-RB',
        'StopPhase-Cmd', 'ChangePhase-Cmd',
        'Log-Mon',
        )

    def __init__(self, devname, properties=None, auto_monitor_mon=True):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in self.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        properties = properties or self._properties + self._properties_papu
        super().__init__(
            devname, properties=properties, auto_monitor_mon=auto_monitor_mon)

    @property
    def period_length(self):
        """Return ID period length [mm]."""
        return self['PeriodLength-Cte']

    @property
    def log_mon(self):
        """Return ID Log."""
        return self['Log-Mon']

    # --- phase speeds ----

    @property
    def phase_speed(self):
        """Return phase speed readback [mm/s]."""
        return self['PhaseSpeed-RB']

    @property
    def phase_speed_mon(self):
        """Return phase speed monitor [mm/s]."""
        return self['PhaseSpeed-Mon']

    @property
    def phase_speed_max(self):
        """Return max phase speed readback [mm/s]."""
        return self['MaxPhaseSpeed-RB']

    @property
    def phase_speed_max_lims(self):
        """."""
        ctrl = self.pv_ctrlvars('MaxPhaseSpeed-SP')
        lims = [ctrl['lower_ctrl_limit'], ctrl['upper_ctrl_limit']]
        return lims

    # --- phase ---

    @property
    def phase_parked(self):
        """Return ID parked phase value [mm]."""
        return self['ParkedPhase-Cte']

    @property
    def phase(self):
        """Return ID phase readback [mm]."""
        return self['Phase-RB']

    @property
    def phase_min(self):
        """Return ID phase lower control limit [mm]."""
        ctrlvars = self.pv_ctrlvars('Phase-SP')
        return ctrlvars['lower_ctrl_limit']

    @property
    def phase_max(self):
        """Return ID phase upper control limit [mm]."""
        ctrlvars = self.pv_ctrlvars('Phase-SP')
        return ctrlvars['upper_ctrl_limit']

    @property
    def phase_mon(self):
        """Return ID phase monitor [mm]."""
        return self['Phase-Mon']

    # --- drive checks ---

    @property
    def is_phase_drive_powered(self):
        """Return phase driver power state on (True|False)."""
        return self['PwrPhase-Mon'] != 0

    # --- movement checks ---

    @property
    def is_move_phase_enabled(self):
        """Return phase movement enabled state (True|False)."""
        return self['AllowedToChangePhase-Mon'] != 0

    @property
    def is_moving(self):
        """Return is moving state (True|False)."""
        return self['Moving-Mon'] != 0

    @property
    def is_homing(self):
        """Return whether ID is in homing procedure (True|False)."""
        return self['Home-Mon'] != 0

    @property
    def is_beamline_ctrl_enabled(self):
        """Return beamline control enabled state (True|False)."""
        return self['BeamLineCtrlEnbl-Sts'] != 0

    # --- cmd_beamline and cmd_drive

    def cmd_drive_turn_power_on(self, timeout=None):
        """Command turn phase drive on."""
        if self.is_phase_drive_powered:
            return True
        self['EnblPwrPhase-Cmd'] = 1
        props_values = {'PwrPhase-Mon': 1}
        return self._wait(props_values, timeout=timeout)

    def cmd_beamline_ctrl_enable(self, timeout=None):
        """Command enable bealine ID control."""
        return self._write_sp('BeamLineCtrlEnbl-Sel', 1, timeout)

    def cmd_beamline_ctrl_disable(self, timeout=None):
        """Command disable bealine ID control."""
        return self._write_sp('BeamLineCtrlEnbl-Sel', 0, timeout)

    # --- cmd_set ---

    def cmd_set_phase(self, phase, timeout=None):
        """Command to set ID target phase for movement [mm]."""
        return self._write_sp('Phase-SP', phase, timeout)

    def cmd_set_phase_speed(self, phase_speed, timeout=None):
        """Command to set ID cruise phase speed for movement [mm/s]."""
        return self._write_sp('PhaseSpeed-SP', phase_speed, timeout)

    def cmd_set_phase_speed_max(self, phase_speed_max, timeout=None):
        """Command to set ID max cruise phase speed for movement [mm/s]."""
        return self._write_sp('MaxPhaseSpeed-SP', phase_speed_max, timeout)

    # --- cmd_move disable/enable ---

    def cmd_move_phase_enable(self, timeout=None):
        """Command to release and enable ID phase movement."""
        return self._write_sp('EnblAndReleasePhase-Sel', 1, timeout)

    def cmd_move_phase_disable(self, timeout=None):
        """Command to disable and break ID phase movement."""
        return self._write_sp('EnblAndReleasePhase-Sel', 0, timeout)

    def cmd_move_enable(self, timeout=None):
        """Command to release and enable ID phase and gap movements."""
        return self.cmd_move_phase_enable(timeout=timeout)

    def cmd_move_disable(self, timeout=None):
        """Command to disable and break ID phase and gap movements."""
        return self.cmd_move_phase_disable(timeout=timeout)

    # --- cmd_wait

    def wait_while_busy(self, timeout=None):
        """Command wait within timeout while ID control is busy."""
        return True

    def wait_move_start(self, timeout=None):
        """Command wait until movement starts or timeout."""
        time_init = _time.time()
        while not self.is_moving:
            if timeout is not None and _time.time() - time_init > timeout:
                return False
        return True

    # -- cmd_move

    def cmd_move_stop(self, timeout=None):
        """Command to interrupt and then enable phase movements."""
        timeout = timeout or self._default_timeout

        # wait for not busy state
        if not self.wait_while_busy(timeout=timeout):
            return False

        # send stop command
        self.cmd_move_disable()

        # check for successful stop
        if not self.wait_while_busy(timeout=timeout):
            return False
        success = True
        success &= super()._wait('Moving-Mon', 0, timeout=timeout)
        if not success:
            return False

        # enable movement again
        return self.cmd_move_enable(timeout=timeout)

    def cmd_move_phase_start(self, timeout=None):
        """Command to start phase movement."""
        return self._move_start('ChangePhase-Cmd', timeout=timeout)

    def cmd_move(self, phase, timeout=None):
        """Command to set and start phase movements."""
        # calc ETA
        dtime_max = abs(phase - self.phase_mon) / self.phase_speed

        # additional percentual in ETA
        tol_dtime = 300  # [%]
        tol_factor = (1 + tol_dtime/100)
        tol_total = tol_factor * dtime_max + 5

        # set target phase and gap
        if not self.cmd_set_phase(phase=phase, timeout=timeout):
            return False

        # command move start
        if not self.cmd_move_phase_start(timeout=timeout):
            return False

        # wait for movement within reasonable time
        time_init = _time.time()
        while \
                abs(self.phase_mon - phase) > PAPU.TOLERANCE_PHASE or \
                self.is_moving:
            if _time.time() - time_init > tol_total:
                print(f'tol_total: {tol_total:.3f} s')
                print(f'wait_time: {_time.time() - time_init:.3f} s')
                print()
                return False
            _time.sleep(self._SHORT_SHUT_EYE)

        # successfull movement at this point
        return True

    def cmd_move_park(self, timeout=None):
        """Command to set and start ID movement to parked config."""
        return self.cmd_move(self.phase_parked, timeout=timeout)

    # --- cmd_reset

    def cmd_device_reset(self, timeout=None):
        """Command to reset ID to a standard movement state."""
        success = True
        success &= self.cmd_beamline_ctrl_disable(timeout=timeout)
        success &= self.cmd_drive_turn_power_on(timeout=timeout)
        success &= self.cmd_move_enable(timeout=timeout)
        return success

    # --- other cmds ---

    def cmd_clear_error(self):
        """."""
        self['ClearErr-Cmd'] = 1

    # --- private methods ---

    def _move_start(self, cmd_propty, timeout=None):
        """."""
        timeout = timeout or self._default_timeout

        # wait for not busy state
        if not self.wait_while_busy(timeout=timeout):
            return False

        # send move command
        self[cmd_propty] = 1

        return True

    def _write_sp(self, propties_sp, values, timeout=None):
        timeout = timeout or self._default_timeout
        if isinstance(propties_sp, str):
            propties_sp = (propties_sp, )
            values = (values, )
        success = True
        for propty_sp, value in zip(propties_sp, values):
            propty_rb = propty_sp.replace('-SP', '-RB').replace('-Sel', '-Sts')
            # if self[propty_rb] == value:
            #     continue
            if not self.wait_while_busy(timeout=timeout):
                return False
            self[propty_sp] = value
            success &= super()._wait(
                propty_rb, value, timeout=timeout, comp='eq')
        return success

    def _wait(self, props_values, timeout=None, comp='eq'):
        timeout = timeout or self._default_timeout
        success = True
        for propty, value in props_values.items():
            success &= super()._wait(
                propty, value, timeout=timeout, comp=comp)
        return success


class EPU(PAPU):
    """EPU Insertion Device."""

    class DEVICES:
        """."""

        EPU50_10SB = 'SI-10SB:ID-EPU50'
        ALL = (EPU50_10SB, )

    _properties = PAPU._properties + (
        'EnblPwrAll-Cmd',
        'PwrGap-Mon',
        'Status-Mon',
        'EnblAndReleaseGap-Sel', 'EnblAndReleaseGap-Sts',
        'AllowedToChangeGap-Mon',
        'ParkedGap-Cte', 'IsBusy-Mon',
        'Gap-SP', 'Gap-RB', 'Gap-Mon',
        'GapSpeed-SP', 'GapSpeed-RB', 'GapSpeed-Mon',
        'MaxGapSpeed-SP', 'MaxGapSpeed-RB',
        'ChangeGap-Cmd', 'Stop-Cmd',
        )

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in EPU.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, properties=self._properties, auto_monitor_mon=True)

    @property
    def status(self):
        """ID status."""
        return self['Status-Mon']

    # --- gap speeds ----

    @property
    def gap_parked(self):
        """Return ID parked gap value [mm]."""
        return self['ParkedGap-Cte']

    @property
    def gap_speed(self):
        """Return gap speed readback [mm/s]."""
        return self['GapSpeed-RB']

    @property
    def gap_speed_mon(self):
        """Return gap speed monitor [mm/s]."""
        return self['GapSpeed-Mon']

    @property
    def gap_speed_max(self):
        """Return max gap speed readback [mm/s]."""
        return self['MaxGapSpeed-RB']

    @property
    def gap_speed_max_lims(self):
        """."""
        ctrl = self.pv_ctrlvars('MaxGapSpeed-SP')
        lims = [ctrl['lower_ctrl_limit'], ctrl['upper_ctrl_limit']]
        return lims

    # --- gap ---

    @property
    def gap(self):
        """Return ID gap readback [mm]."""
        return self['Gap-RB']

    @property
    def gap_min(self):
        """Return ID gap lower control limit [mm]."""
        ctrlvars = self.pv_ctrlvars('Gap-SP')
        return ctrlvars['lower_ctrl_limit']

    @property
    def gap_max(self):
        """Return ID gap upper control limit [mm]."""
        ctrlvars = self.pv_ctrlvars('Gap-SP')
        return ctrlvars['upper_ctrl_limit']

    @property
    def gap_mon(self):
        """Return ID gap monitor [mm]."""
        return self['Gap-Mon']

    # --- drive checks ---

    @property
    def is_gap_drive_powered(self):
        """Return gap driver power state on (True|False)."""
        return self['PwrGap-Mon'] != 0

    @property
    def is_drives_powered(self):
        """Return phase & gap drives powered state on (True|False)."""
        return self.is_phase_drive_powered and self.is_gap_drive_powered

    # --- movement checks ---

    @property
    def is_move_gap_enabled(self):
        """Return gap movement enabled state (True|False)."""
        return self['AllowedToChangeGap-Mon'] != 0

    @property
    def is_move_enabled(self):
        """Return phase and gap movements enabled state (True|False)."""
        return self.is_move_phase_enabled and self.is_move_gap_enabled

    @property
    def is_homing(self):
        """Return whether ID is in homing procedure."""
        return False

    # --- other checks ---

    @property
    def is_busy(self):
        """Return is busy state (True|False)."""
        return self['IsBusy-Mon'] != 0

    # --- cmd_beamline and cmd_drive

    def cmd_drive_turn_power_on(self, timeout=None):
        """Command turn phase and gap drives on."""
        if self.is_phase_drive_powered and self.is_gap_drive_powered:
            return True
        self['EnblPwrAll-Cmd'] = 1
        props_values = {'PwrPhase-Mon': 1, 'PwrGap-Mon': 1}
        return self._wait(props_values, timeout=timeout)

    # --- cmd_set ---

    def cmd_set_gap(self, gap, timeout=None):
        """Command to set ID target gap for movement [mm]."""
        return self._write_sp('Gap-SP', gap, timeout)

    def cmd_set_gap_speed(self, gap_speed, timeout=None):
        """Command to set ID cruise gap speed for movement [mm/s]."""
        return self._write_sp('GapSpeed-SP', gap_speed, timeout)

    def cmd_set_gap_speed_max(self, gap_speed_max, timeout=None):
        """Command to set ID max cruise gap speed for movement [mm/s]."""
        return self._write_sp('MaxGapSpeed-SP', gap_speed_max, timeout)

    # --- cmd_move disable/enable ---

    def cmd_move_gap_enable(self, timeout=None):
        """Command to release and enable ID gap movement."""
        return self._write_sp('EnblAndReleaseGap-Sel', 1, timeout)

    def cmd_move_gap_disable(self, timeout=None):
        """Command to disable and break ID gap movement."""
        return self._write_sp('EnblAndReleaseGap-Sel', 0, timeout)

    def cmd_move_enable(self, timeout=None):
        """Command to release and enable ID phase and gap movements."""
        success = True
        success &= self.cmd_move_phase_enable(timeout=timeout)
        success &= self.cmd_move_gap_enable(timeout=timeout)
        return success

    def cmd_move_disable(self, timeout=None):
        """Command to disable and break ID phase and gap movements."""
        success = True
        success &= self.cmd_move_phase_disable(timeout=timeout)
        success &= self.cmd_move_gap_disable(timeout=timeout)
        return success

    # --- cmd_wait

    def wait_while_busy(self, timeout=None):
        """Command wait within timeout while ID control is busy."""
        timeout = timeout or self._default_timeout
        time_init = _time.time()
        while self.is_busy:
            _time.sleep(min(self._SHORT_SHUT_EYE, timeout))
            if _time.time() - time_init > timeout:
                return False
        return True

    # -- cmd_move

    def cmd_move_gap_start(self, timeout=None):
        """Command to start gap movement."""
        return self._move_start('ChangeGap-Cmd', timeout=timeout)

    def cmd_move(self, phase, gap, timeout=None):
        """Command to set and start phase and gap movements."""
        # calc ETA
        dtime_phase = abs(phase - self.phase_mon) / self.phase_speed
        dtime_gap = abs(gap - self.gap_mon) / self.gap_speed
        dtime_max = max(dtime_phase, dtime_gap)

        # additional percentual in ETA
        tol_gap = 0.01  # [mm]
        tol_phase = 0.01  # [mm]
        tol_dtime = 300  # [%]
        tol_factor = (1 + tol_dtime/100)
        tol_total = tol_factor * dtime_max + 5

        # set target phase and gap
        if not self.cmd_set_phase(phase=phase, timeout=timeout):
            return False
        if not self.cmd_set_gap(gap=gap, timeout=timeout):
            return False

        # command move start
        if not self.cmd_move_phase_start(timeout=timeout):
            return False
        if not self.cmd_move_gap_start(timeout=timeout):
            return False

        # wait for movement within reasonable time
        time_init = _time.time()
        while \
                abs(self.gap_mon - gap) > tol_gap or \
                abs(self.phase_mon - phase) > tol_phase or \
                self.is_moving:
            if _time.time() - time_init > tol_total:
                print(f'tol_total: {tol_total:.3f} s')
                print(f'wait_time: {_time.time() - time_init:.3f} s')
                print()
                return False
            _time.sleep(EPU._SHORT_SHUT_EYE)

        # successfull movement at this point
        return True

    def cmd_move_park(self, timeout=None):
        """Command to set and start ID movement to parked config."""
        return self.cmd_move(
            self.phase_parked, self.gap_parked, timeout=timeout)

    # --- other cmds ---

    def cmd_clear_error(self):
        """."""
        pass


class WIG(_Device):
    """Wiggler Insertion Device."""

    class DEVICES:
        """."""
        WIG180_14SB = 'SI-14SB:ID-WIG180'
        ALL = (WIG180_14SB, )

    # NOTE: IOC yet to be written!
    _properties = (
        'Gap-SP', 'Gap-RB', 'Gap-Mon',
    )

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in WIG.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, properties=WIG._properties, auto_monitor_mon=True)



"""BSMP protocol implementation."""
import typing

from . import constants as _const
from .entities import Entities as _Entities
from .exceptions import SerialAnomResp as _SerialAnomResp
from .exceptions import SerialError as _SerialError
from .serial import Channel as _Channel
from .serial import IOInterface as _IOInterface
from .serial import Message as _Message


class BSMP:
    """BSMP protocol implementation for Master Node."""

    _timeout_execute_function: float = 100.0  # [ms]

    def __init__(self, iointerf: _IOInterface, slave_address: int, entities: _Entities):
        """Constructor."""
        self._entities: _Entities = entities
        self._channel: _Channel = _Channel(iointerf, slave_address)

    @property
    def entities(self) -> _Entities:
        """Return BSMP entities."""
        return self._entities

    @property
    def channel(self) -> _Channel:
        """Return serial channel to an address."""
        return self._channel

    # 0x0_
    def query_protocol_version(self):
        """Consult protocol version. Command 0x00."""
        # TODO: needs implementation!
        raise NotImplementedError()

    def query_list_of_variables(self):
        """Consult list of variables. Command 0x02."""
        # TODO: needs implementation!
        raise NotImplementedError()

    def query_list_of_group_of_variables(self, timeout: float) -> typing.Tuple[int, typing.Optional[typing.List[typing.Tuple[bool, int]]]]:
        """Consult groups list. Command 0x04."""
        # command and expected response
        cmd, ack = _const.CMD_QUERY_LIST_OF_GROUP_OF_VARIABLES, \
            _const.CMD_LIST_OF_GROUP_OF_VARIABLES

        # build payload
        payload: typing.List[str] = []

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        # expected response
        if res.cmd == ack:
            groupdata: typing.List[typing.Tuple[bool, int]] = []
            for groupchar in res.payload:
                byte = ord(groupchar)
                waccess = (byte & 0b10000000) > 0
                nrvars = (byte & 0b01111111)
                groupdata.append((waccess, nrvars))
            return _const.ACK_OK, groupdata

        # anomalous response
        return BSMP.anomalous_response(cmd, res.cmd)

    def query_group_of_variables(
        self,
        group_id: int,
        timeout: float
    ) -> typing.Tuple[int, typing.Optional[typing.List[int]]]:
        """Return id of the variables in the given group."""
        # command and expected response
        cmd, ack = _const.CMD_QUERY_GROUP_OF_VARIABLES, \
            _const.CMD_GROUP_OF_VARIABLES

        # build payload
        payload = [chr(group_id)]

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        # expected response
        if res.cmd == ack:
            return _const.ACK_OK, list(map(ord, res.payload))

        # anomalous response
        return BSMP.anomalous_response(cmd, res.cmd)

    def query_list_of_curves(self):
        """Consult curves_list. Command 0x08."""
        raise NotImplementedError()

    def query_curve_checksum(self, curve_id: int):
        """Consult curve checksum given curve id. Command 0x0A."""
        raise NotImplementedError()

    def query_list_of_functions(self):
        """Consult functions list. Command 0x0C."""
        raise NotImplementedError()

    # 0x1_
    def read_variable(
        self,
        var_id: int,
        timeout: float
    ) -> typing.Union[typing.Tuple[None, None], typing.Tuple[int, typing.Any]]:
        """Read variable."""
        # command and expected response
        cmd, ack = _const.CMD_READ_VARIABLE, _const.CMD_VARIABLE_VALUE

        # build payload
        payload = [chr(var_id)]

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        if res.cmd == ack:
            variable = self.entities.variables[var_id]
            if len(res.payload) == variable.size:
                # expected response
                return _const.ACK_OK, variable.load_to_value(res.payload)

            # unexpected variable size
            fmts = 'Unexpected BSMP variable size for command 0x{:02X}: {}!'
            print(fmts.format(cmd, res.cmd))
            return None, None

        # anomalous response
        return BSMP.anomalous_response(
            cmd, res.cmd, var_id=var_id, payload=res.payload)

    def read_group_of_variables(
        self,
        group_id: int,
        timeout: float
    ):
        """Read variable group."""
        # command and expected response
        cmd, ack = \
            _const.CMD_READ_GROUP_OF_VARIABLES, \
            _const.CMD_GROUP_OF_VARIABLES_VALUE

        # build payload
        payload = [chr(group_id)]

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        if res.cmd == ack:
            # expected response
            group = self.entities.groups[group_id]
            if len(res.payload) == group.variables_size():
                return _const.ACK_OK, group.load_to_value(res.payload)
            # unexpected group variables size
            return BSMP.anomalous_response(
                cmd, res.cmd,
                group_id=group_id,
                payload_len=len(res.payload),
                var_size=group.variables_size()
            )

        # anomalous response
        return BSMP.anomalous_response(cmd, res.cmd)

    # 0x2_
    def write_variable(self, var_id, value):
        """Write to variable. Command 0x20."""
        # TODO: needs implementation!
        raise NotImplementedError()

    def write_group_of_variables(self, group_id, value):
        """Write to the variables of a group. Command 0x22."""
        # TODO: needs implementation!
        raise NotImplementedError()

    def binoperation_variable(self, var_id, operation, mask):
        """Perform a binary operation to variable. Command 0x24."""
        # TODO: needs implementation!
        raise NotImplementedError()

    def binoperation_group(self, group_id, operation, mask):
        """Perform a binary oeration to the vars of a group. Command 0x26."""
        # TODO: needs implementation!
        raise NotImplementedError()

    def write_and_read_variable(self, w_var_id, r_var_id, value):
        """Write value to a var and read another var value. Command 0x28."""
        # TODO: needs implementation!
        raise NotImplementedError()

    # 0x3_
    def create_group_of_variables(
        self,
        var_ids: typing.List[int],
        timeout: float
    ) -> typing.Tuple[typing.Optional[int], typing.Optional[typing.List[str]]]:
        """Create new group with given variable ids."""
        cmd, ack = \
            _const.CMD_CREATE_GROUP_OF_VARIABLES, _const.ACK_OK

        # build payload
        var_ids = sorted(var_ids)
        payload = [chr(var_id) for var_id in var_ids]

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        if res.cmd == ack:
            if not res.payload:
                # expected response
                # TODO: revise this. do we need to change entities?!
                # if so, should we make a local copy instead?
                self.entities.add_group(var_ids)
                return _const.ACK_OK, None
            # unexpected non-empty response payload
            fmts = ('Unexpected BSMP non-empty resp payload '
                    'for command 0x{:02X}: {}!')
            print(fmts.format(cmd, res.cmd))
            return None, None

        # anomalous response
        return BSMP.anomalous_response(cmd, res.cmd)

    def remove_all_groups_of_variables(
        self,
        timeout: float
    ) -> typing.Tuple[int, typing.Optional[typing.List[str]]]:
        """Remove all groups."""
        cmd, ack = \
            _const.CMD_REMOVE_ALL_GROUPS_OF_VARIABLES, _const.ACK_OK

        # build payload
        payload: typing.List[str] = []

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        if res.cmd == ack and not res.payload:
            # expected response
            # TODO: revise this. do we need to change entities?!
            # if so, should we make a local copy instead?
            self.entities.remove_all_groups_of_variables()
            return _const.ACK_OK, None

        # anomalous response
        return BSMP.anomalous_response(cmd, res.cmd, payload=res.payload)

    # 0x4_
    def request_curve_block(
            self,
            curve_id,
            block,
            timeout: float,
            print_error: bool = True
    ) -> typing.Tuple[typing.Optional[int], typing.Optional[typing.List[str]]]:
        """Read curve block."""
        # command and expected response
        cmd, ack = _const.CMD_REQUEST_CURVE_BLOCK, _const.CMD_CURVE_BLOCK

        # build payload
        lsb, hsb = block & 0xff, (block & 0xff00) >> 8
        payload = [chr(curve_id), chr(hsb), chr(lsb)]

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        if res.cmd == ack:
            data = res.payload[3:]
            curve = self.entities.curves[curve_id]
            if len(data) % curve.type.size:
                # unexpected curve size
                fmts = ('Curve size is not multiple of curve.type.size!\n'
                        ' received curce size: {}\n'
                        ' curve.type.size: {}')
                print(fmts.format(len(data), curve.type.size))
                return None, None
            cid = ord(res.payload[0])
            cblock = (ord(res.payload[1]) << 8) + ord(res.payload[2])
            if cid != curve_id or cblock != block:
                # unexpected curve id or block number
                fmts = ('Invalid curve id or block offset in response!\n'
                        ' expected - curve_id:{}, block_offset:{}\n'
                        ' received - curve_id:{}, block_offset:{}')
                print(fmts.format(curve_id, block, cid, cblock))
                return None, None

            # expected result
            data_float = curve.load_to_value(data)
            return _const.ACK_OK, data_float

        # anomalous response
        return BSMP.anomalous_response(cmd, res.cmd, print_error=print_error)

    def curve_block(
            self,
            curve_id: int,
            block,
            value,
            timeout: float
    ) -> typing.Tuple[int, typing.Optional[typing.List[str]]]:
        """Write to curve block."""
        # command and expected response
        cmd, ack = _const.CMD_CURVE_BLOCK, _const.ACK_OK

        # build payload
        curve = self.entities.curves[curve_id]
        lsb, hsb = block & 0xff, (block & 0xff00) >> 8
        payload = [chr(curve_id), chr(hsb), chr(lsb)] + \
            curve.value_to_load(value)

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        if res.cmd == ack:
            # expected response
            return res.cmd, res.payload

        # anomalous response
        return BSMP.anomalous_response(cmd, res.cmd)

    def recalculate_curve_checksum(
        self,
        curve_id: int,
        timeout: float
    ) -> typing.Tuple[int, typing.Optional[typing.List[str]]]:
        """Recalculate curve checksum."""
        # command and expected response
        cmd, ack = \
            _const.CMD_RECALCULATE_CURVE_CHECKSUM, _const.CMD_CURVE_CHECKSUM

        # build payload
        payload = [chr(curve_id)]

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        if res.cmd == ack:
            # expected response
            return res.cmd, res.payload

        # anomalous response
        return BSMP.anomalous_response(cmd, res.cmd)

    # 0x5_
    def execute_function(
        self,
        func_id: int,
        input_val=None,
        timeout: float = _timeout_execute_function,
        read_flag: bool = True,
        print_error: bool = True
    ) -> typing.Optional[typing.Tuple[int, typing.Optional[typing.Union[typing.List[str], str]]]]:
        """Execute a function.

        parameter:
            read_flag: whether to execute a read after a write.
        """
        # command and expected response
        cmd, ack = _const.CMD_EXECUTE_FUNCTION, _const.CMD_FUNCTION_RETURN

        # build payload
        function = self.entities.functions[func_id]
        payload = [chr(func_id)] + function.value_to_load(input_val)

        # send request package
        try:
            msg = _Message.message(cmd, payload=payload)
            res = self.channel.request(msg, timeout=timeout)
        except (TimeoutError, ValueError) as err:
            raise _SerialError(err)

        # TODO: This should be temporary. It is used for ps F_RESET_UDC.
        # PS Firmware should change so that F_RESET_UDC returns ack.
        if not read_flag:
            return None

        if res.cmd == ack:
            if len(res.payload) == function.o_size:
                # expected response
                return _const.ACK_OK, function.load_to_value(res.payload)
        if res.cmd == _const.CMD_FUNCTION_ERROR:
            # function error
            if len(res.payload) == 1:
                # return error code
                return res.cmd, res.payload[0]

        # anomalous response
        return BSMP.anomalous_response(
            cmd, res.cmd, func_id=func_id, print_error=print_error)

    @staticmethod
    def anomalous_response(cmd, ack: int, **kwargs) -> typing.Tuple[int, None]:
        """Print information about anomalous response."""
        # response with error
        if _const.ACK_OK < ack <= _const.ACK_RESOURCE_BUSY:
            if 'print_error' not in kwargs or kwargs['print_error']:
                fmts = 'BSMP response (error) for command 0x{:02X}: 0x{:02X}!'
                print(fmts.format(cmd, ack))
                for key, value in kwargs.items():
                    print('{}: {}'.format(key, value))
            return ack, None

        # unexpected response, raise Exception
        fmts = 'BSMP response (unexpected) for command 0x{:02X}: 0x{:02X}!'
        errmsg = fmts.format(cmd, ack)
        if 'print_error' not in kwargs or kwargs['print_error']:
            print(errmsg)
            for key, value in kwargs.items():
                print('{}: {}'.format(key, value))
        raise _SerialAnomResp(errmsg)
